from pwn import *
import sys

def repr_hexarr(arr, sz, name=''):
    s = ''
    if name != '':
        s += f'{name}:'
    for x in arr:
        if sz == 4:
            s += f' {x:08x}'
        elif sz == 2:
            s += f' {x:04x}'
        elif sz == 1:
            s += f' {x:02x}'
    return s

class Instruction:
    # 0x56 (V), 0x20 ( ), 0x9, 0x1
    ALU_OP_MAP = ['RL5', 'XOR', 'ADD', 'AND']
    OPCODE_MAP = ['SYS', 'B']

    def __init__(self, ins, off):
        self._ins = ins
        
        self.addr = off
        self.opcode = (ins & 3)
        self.alu_op = (ins >> 2) & 3
        self.dst = (ins >> 4) & 3
        self.src = (ins >> 6) & 3
        self.cond = (ins >> 2) & 1
        self.target = (ins >> 3) & 0x1f
    
    def __repr__(self):
        s = f'{self.addr:02x} ({self._ins:02x} {self._ins:08b}): '
        if self.opcode & 1 == 0:
            s += f'{self.OPCODE_MAP[(self.opcode >> 1) & 1]}'
            if self.opcode == 2:
                if self.cond == 0:
                    s += '   '
                else:
                    s += 'Z  '
            else:
                s += ' '
        else:
            s += f'{self.ALU_OP_MAP[self.alu_op]} '
        s += f'R{self.dst} '
        if self.opcode & 2:
            s += f'I{self.src} '
        else:
            s += f'R{self.src} '
        s += f'T {(self.target) % 32:02x}\t'

        s += '['
        s += 'S' if self._ins & 0b100 else ' '
        s += '<' if self._ins & 0b1000 else ' '
        s += '>' if self._ins & 0b10000 else ' '
        s += 'F' if self._ins & 0b100000 else ' '
        s += 'M' if self._ins & 0b1000000 else ' '
        s += 'E' if self._ins & 0b10000000 else ' '
        s += ']'
        return s

class ProgramHeader:
    def __init__(self, b):
        self._b = b

        self.iterations = u32(b[0:4])
        self.timeout = u32(b[4:8])
        self.io_out_gate = u32(b[8:0xc])
        self.io_out_byte = []
        for i in range(8):
            self.io_out_byte.append(u32(b[0xc+4*i:0xc+4*i+4]))
        self.io_in_gate = u32(b[0x2c:0x30])
        self.io_in_byte = []
        for i in range(8):
            self.io_in_byte.append(u32(b[0x30+4*i:0x30+4*i+4]))
        self.memtest = u32(b[0x50:0x54])
        self.exit_pin = u32(b[0x54:0x58])
        self.flag_pin = u32(b[0x58:0x5c])
        self.memaddr = []
        for i in range(5):
            self.memaddr.append(u32(b[0x5c+4*i:0x5c+4*i+4]))
        self.membyte = []
        for i in range(8):
            self.membyte.append(u32(b[0x70+4*i:0x70+4*i+4]))
        self.mem = [Instruction(c, i) for i, c in enumerate(b[0x90:0xb0])]
    
    def __repr__(self):
        s = f'iterations: {self.iterations:08x}\n'
        s += f'timeout: {self.timeout}s\n'
        s += f'io_out_gate: {self.io_out_gate:08x}\n'
        s += repr_hexarr(self.io_out_byte, 4, 'io_out_byte') + '\n'
        s += f'io_in_gate: {self.io_in_gate:08x}\n'
        s += repr_hexarr(self.io_in_byte, 4, 'io_in_byte') + '\n'
        s += f'memtest: {self.memtest:08x}\n'
        s += f'exit_pin: {self.exit_pin:08x}\n'
        s += f'flag_pin: {self.flag_pin:08x}\n'
        s += repr_hexarr(self.memaddr, 4, 'memaddr') + '\n'
        s += repr_hexarr(self.membyte, 4, 'membyte') + '\n'
        s += f'mem:\n'
        #s += f'  {Instruction(0, 0xff)}\n'
        for ins in self.mem:
            s += f'  {ins}\n'
        return s

class ProgramGate:
    GATE_HEAP = 0x4040000

    # CTYPE_DYN  = 1
    # CTYPE_INV = 2
    # CTYPE_NOP = 4
    # CTYPE_REG = 5
    # CTYPE_06  = 6

    TYPE_CMP = 0
    TYPE_DYN = 1
    TYPE_MEM = 3
    TYPE_NOP = 4
    TYPE_REG = 5
    TYPE_CHK = 0x14

    TYPE_MAP = {
        TYPE_CMP: 'CMP',
        TYPE_DYN: 'DYN',
        TYPE_MEM: 'MEM',
        TYPE_NOP: 'NOP',
        TYPE_REG: 'REG',
        TYPE_CHK: 'CHK'
    }

    CHECK_DO_CHECK = 0x2 << 3
    CHECK_IS_ARCH_CHECK = 0x4 << 3

    def __init__(self, b, off):
        self._b = b

        self.addr = self.GATE_HEAP + off
        self.type = b[0] & 0x1f
        #self.do_check = (b[0] & self.CHECK_DO_CHECK) != 0
        self.is_arch_check = (b[0] & self.CHECK_IS_ARCH_CHECK) != 0
        self.perturb = b[1]
        self.pac = u16(b[2:4])
        self.pac_i = u32(b[4:8])
        self.pac_c = u32(b[8:0xc])
        self.pacfail = u32(b[0xc:0x10])
        self.pacpass = u32(b[0x10:0x14])
        self.out = u32(b[0x14:0x18])
        #self.usedby = u32(b[0x18:0x1c])
        self.usedby = []

        self.io_out_gate = False
        self.io_out_byte = -1
        self.io_in_gate = False
        self.io_in_byte = -1
        self.memtest = False
        self.exit_pin = False
        self.flag_pin = False
        self.memaddr = -1
        self.membyte = -1

    def __repr__(self):
        s = f'{self.addr:08x}: {self.TYPE_MAP[self.type]} '

        if self.type == self.TYPE_CMP:
            s += f'{self.pac:04x} != '
            s += f'PAC([{self.pac_i:08x}], [{self.pac_c:08x}]'
            if self.perturb:
                s += f'+{self.perturb:02x}h'
            s += f') ? '
            s += f'{self.pacfail:08x} : {self.pacpass:08x}'
            assert(self.out == 0)
        elif self.type == self.TYPE_DYN:
            assert(False)
        elif self.type == self.TYPE_MEM:
            assert(self.membyte != -1)
            s += f'B{self.membyte} (0: {self.pacpass:08x} 1: {self.pacfail:08x}) INIT: {0 if self.out == self.pacpass else 2} ({self.out:08x})'
        elif self.type == self.TYPE_NOP:
            s += f'[{self.pac_i:08x}] (0: {self.pacpass:08x} 1: {self.pacfail:08x} A: {self.is_arch_check})'
            if self.io_out_gate:
                s += f' [IO_OUT]'
            if self.io_in_gate:
                s += f' [IO_IN]'
            if self.memtest:
                s += f' [MEMTEST (MUST BE 0)]'
            if self.exit_pin:
                s += f' [EXIT IF 0]'
            if self.flag_pin:
                s += f' [FLAG]'
            if self.memaddr != -1:
                s += f' [MEMADDR_B{self.memaddr}]'
        elif self.type == self.TYPE_REG:
            s += f'[{self.pac_i:08x}] (0: {self.pacpass:08x} 1: {self.pacfail:08x}) INIT: {0 if self.out == self.pacpass else 1} ({self.out:08x})'
            if self.io_out_byte != -1:
                s += f' [IO_OUT_B{self.io_out_byte}]'
            if self.io_in_byte != -1:
                s += f' [IO_IN_B{self.io_in_byte}]'
        elif self.type == self.TYPE_CHK:
            s += f'ASSERT([{self.pac_i:08x}] == {self.pacfail:08x} A: {self.is_arch_check}) '

        #s = f'{self.addr:08x}: {self.TYPE_MAP[self.type]} P{self.perturb:02x} PAC {self.pac:04x} I {self.pac_i:08x} C {self.pac_c:08x} PF {self.pacfail:08x} PP {self.pacpass:08x} O {self.out:08x}'

        #s += ' U' + repr_hexarr(self.usedby, 4)
        return s

class Program:
    def __init__(self, b):
        self._b = b

        self.header = ProgramHeader(b[:0xb0])
        self.gates = {}
        for i in range(0, len(b[0xb0:]), 0x1c):
            g = ProgramGate(b[0xb0+i:0xb0+i+0x1c], i)
            self.gates[g.addr] = g
        
        self.preprocess()

    def preprocess(self):
        for addr, g in self.gates.items():
            if g.pac_i != 0:
                self.gates[g.pac_i].usedby.append(addr)
            if g.pac_c != 0:
                self.gates[g.pac_c].usedby.append(addr)
        
        self.gates[self.header.io_out_gate].io_out_gate = True
        for i, addr in enumerate(self.header.io_out_byte):
            self.gates[addr].io_out_byte = i
        self.gates[self.header.io_in_gate].io_in_gate = True
        for i, addr in enumerate(self.header.io_in_byte):
            self.gates[addr].io_in_byte = i
        self.gates[self.header.memtest].memtest = True
        self.gates[self.header.exit_pin].exit_pin = True
        self.gates[self.header.flag_pin].flag_pin = True
        for i, addr in enumerate(self.header.memaddr):
            self.gates[addr].memaddr = i
        for i, addr in enumerate(self.header.membyte):
            self.gates[addr].membyte = i
        
    def __repr__(self):
        s = f'Header:\n'
        s += f'{self.header}\n'
        s += f'Gates:\n'
        for addr in sorted(self.gates.keys()):
            s += f'{self.gates[addr]}\n'

        return s

if __name__ == '__main__':
    with open(sys.argv[1], 'rb') as f:
        prog_bytes = f.read()
    
    prog = Program(prog_bytes)
    print(prog)